import abc
import json
from typing import TypeVar, Generator, Generic, Type

from amino.tc.base import TypeClass
from amino import Either, L, _, Map
from amino.do import tdo
from amino.json.data import JsonError, Json, JsonObject, JsonScalar
from amino.util.tpe import qualified_type

A = TypeVar('A')


class Encoder(Generic[A], TypeClass):

    @abc.abstractmethod
    def encode(self, a: A) -> Either[JsonError, Json]:
        ...


@tdo(Either[JsonError, Json])
def encode_json(data: A) -> Generator:
    enc = yield Encoder.e_for(data).lmap(L(JsonError)(data, _))
    yield enc.encode(data)


def dump_json(data: A) -> Either[JsonError, str]:
    return encode_json(data) / _.native / json.dumps


def json_object_with_type(fields: Map[str, JsonObject], tpe: Type[A]) -> Json:
    return JsonObject(fields.cat(('__type__', JsonScalar(qualified_type(tpe)))))

__all__ = ('Encoder', 'encode_json', 'dump_json')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0xd41c3ed4

# Compiled with Coconut version 1.3.0 [Dead Parrot]

# Coconut Header: -------------------------------------------------------------

from __future__ import generator_stop
import sys as _coconut_sys, os.path as _coconut_os_path
_coconut_file_path = _coconut_os_path.dirname(_coconut_os_path.abspath(__file__))
_coconut_sys.path.insert(0, _coconut_file_path)
from __coconut__ import _coconut, _coconut_NamedTuple, _coconut_MatchError, _coconut_tail_call, _coconut_tco, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_pipe, _coconut_star_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial
from __coconut__ import *
_coconut_sys.path.remove(_coconut_file_path)

# Compiled Coconut: -----------------------------------------------------------

from typing import TypeVar
from typing import Generic
from typing import Callable
from typing import Union
from typing import Any
from typing import cast
from typing import Optional
from typing import Type
from typing import Iterator
from typing import List as TList
from typing import Awaitable
from functools import wraps
from operator import eq
import inspect

from amino import boolean
from amino.tc.base import F
from amino.func import call_by_name
from amino.func import I
from amino.func import curried
from amino.func import CallByName

A = TypeVar('A')
B = TypeVar('B')


class Maybe(Generic[A], F[A], implicits=True):

    __slots__ = ()

    @_coconut_tco
    def __new__(cls, value: 'Optional[A]') -> "'Maybe[A]'":
        return _coconut_tail_call(Maybe.check, value)

    @staticmethod
    def optional(value: 'Optional[A]') -> "'Maybe[A]'":
        return Nothing if value is None else Just(value)

    @staticmethod
    @_coconut_tco
    def check(value: 'Optional[A]') -> "'Maybe[A]'":
        return _coconut_tail_call(Maybe.optional, value)

    @staticmethod
    def typed(value: 'Union[A, B]', tpe: 'Type[A]') -> "'Maybe[A]'":
        return Just(value) if isinstance(value, tpe) else Nothing

    @staticmethod
    def wrap(mb: "Union['Maybe[A]', None]") -> "'Maybe[A]'":
        return mb if mb is not None and isinstance(mb, Just) else Nothing

    @staticmethod
    def getattr(obj: 'Any', attr: 'str') -> "'Maybe[A]'":
        return Just(getattr(obj, attr)) if hasattr(obj, attr) else Nothing

    @staticmethod
    @curried
    def iff(cond: 'bool', a: 'Union[A, Callable[[], A]]') -> "'Maybe[A]'":
        return cast(Maybe, Just(call_by_name(a))) if cond else Nothing

    @staticmethod
    @curried
    def iff_m(cond: 'bool', a: "Union['Maybe[A]', Callable[[], 'Maybe[A]']]") -> "'Maybe[A]'":
        return cast(Maybe, call_by_name(a)) if cond else Nothing

    @property
    def _get(self) -> 'Union[A, None]':
        pass

    def cata(self, f: 'Callable[[A], B]', b: 'Union[B, Callable[[], B]]') -> 'B':
        return (f(cast(A, self._get)) if self.is_just else call_by_name(b))

    @_coconut_tco
    def filter(self, f: 'Callable[[A], bool]') -> "'Maybe[A]'":
        return _coconut_tail_call(self.flat_map, lambda a: self if f(a) else Nothing)

    @_coconut_tco
    def get_or_else(self, a: 'Union[A, Callable[[], A]]') -> 'A':
        return _coconut_tail_call(self.cata, cast(Callable, I), a)

    __or__ = get_or_else

    @_coconut_tco
    def get_or_raise(self, e: 'CallByName') -> 'A':
        def raise_e() -> 'None':
            raise call_by_name(e)
        return _coconut_tail_call(self.cata, cast(Callable, I), raise_e)

    @_coconut_tco
    def get_or_fail(self, err: 'CallByName') -> 'A':
        return _coconut_tail_call(self.get_or_raise, lambda: Exception(call_by_name(err)))

    @_coconut_tco
    def __contains__(self, v: 'A') -> 'boolean.Boolean':
        return _coconut_tail_call(self.contains, v)

    def error(self, f: 'Callable[[], Any]') -> "'Maybe[A]'":
        self.cata(cast(Callable, I), f)
        return self

    def observe(self, f: 'Callable[[A], Any]') -> "'Maybe[A]'":
        self.foreach(f)
        return self

    effect = observe

    @_coconut_tco
    def __iter__(self) -> 'Iterator':
        return _coconut_tail_call(iter, self.to_list)

    @property
    @_coconut_tco
    def is_just(self) -> 'boolean.Boolean':
        return _coconut_tail_call(boolean.Boolean, isinstance(self, Just))

    @property
    def is_empty(self) -> 'boolean.Boolean':
        return ~self.is_just

    empty = is_empty

    @property
    @_coconut_tco
    def to_list(self) -> 'TList[A]':
        from amino.list import List
        from amino.list import Nil
        return _coconut_tail_call(self.cata, List, Nil)

    @property
    async def unsafe_await(self) -> "'Maybe[Awaitable]'":
        if self.is_just:
            ret = await cast(Callable[[], Awaitable], self._get)()
            return Maybe(ret)
        else:
            return cast(Maybe[Awaitable], self)

    @property
    @_coconut_tco
    def contains_coro(self) -> 'boolean.Boolean':
        return _coconut_tail_call(self.exists, inspect.iscoroutine)

    @property
    @_coconut_tco
    def json_repr(self) -> 'Optional[A]':
        return _coconut_tail_call(self.cata, cast(Callable, I), lambda: None)


class Just(_coconut_NamedTuple("Just", [("x", 'A')]), Generic[A], Maybe[A]):
    __slots__ = ()
    __ne__ = _coconut.object.__ne__
    @property
    def _get(self) -> 'Optional[A]':
        return self.x

    @_coconut_tco
    def __str__(self) -> 'str':
        return _coconut_tail_call('Just({!s})'.format, self.x)

    @_coconut_tco
    def __repr__(self) -> 'str':
        return _coconut_tail_call('Just({!r})'.format, self.x)

    @_coconut_tco
    def __hash__(self) -> 'int':
        return _coconut_tail_call(hash, self.x)


class _Nothing(Generic[A], Maybe[A]):

    __instance: 'Optional[_Nothing]' = None

    @_coconut_tco
    def __new__(cls: "Type['_Nothing']", *args: 'Any', **kwargs: 'Any') -> "'_Nothing[A]'":
        if _Nothing.__instance is None:
            _Nothing.__instance = object.__new__(cls)
        return _coconut_tail_call(cast, _Nothing, _Nothing.__instance)

    def __str__(self) -> 'str':
        return 'Nothing'

    __repr__ = __str__

    @_coconut_tco
    def __eq__(self, other: 'Any') -> 'bool':
        return _coconut_tail_call(isinstance, other, _Nothing)

    @_coconut_tco
    def __hash__(self) -> 'int':
        return _coconut_tail_call(hash, 'Nothing')

Empty = _Nothing
Nothing: Maybe = _Nothing()


def may(f: 'Callable[..., Optional[A]]') -> 'Callable[..., Maybe[A]]':
    @wraps(f)
    @_coconut_tco
    def wrapper(*a: 'Any', **kw: 'Any') -> 'Maybe[A]':
        return _coconut_tail_call(Maybe.check, f(*a, **kw))
    return wrapper


def flat_may(f: 'Callable[..., Maybe[A]]') -> 'Callable[..., Maybe[A]]':
    @wraps(f)
    def wrapper(*a: 'Any', **kw: 'Any') -> 'Maybe[A]':
        res = f(*a, **kw)
        return res if isinstance(res, Maybe) else Nothing
    return wrapper

__all__ = ('Maybe', 'Just', 'Nothing')



import abc
from typing import TypeVar, Any, Type, Generic

from amino.tc.base import TypeClass
from amino import Either, Map, Right, Lists, Do, _, L, Left
from amino.do import do
from amino.json.data import JsonError, JsonObject, JsonArray, JsonScalar, Json, JsonAbsent
from amino.json.parse import parse_json
from amino.dispatch import dispatch_alg

A = TypeVar('A')


class Decoder(Generic[A], TypeClass):

    @abc.abstractmethod
    def decode(self, tpe: Type[A], data: Map[str, Any]) -> Either[JsonError, A]:
        ...


@do(Either[str, A])
def decode_json_object(data: dict) -> Do:
    m = Map(data)
    tpe_s = yield m.get('__type__').to_either(f'no `__type__` attr in json object {m}')
    tpe = yield Either.import_path(tpe_s)
    dec = yield Decoder.e(tpe)
    yield dec.decode(tpe, m)


class Decode:

    @do(Either[JsonError, A])
    def decode_json_object(self, json: JsonObject) -> Do:
        tpe = yield json.tpe
        dec = yield Decoder.e(tpe).lmap(L(JsonError)(json, _))
        yield dec.decode(tpe, json)

    def decode_json_array(self, json: JsonArray) -> Either[JsonError, A]:
        return Lists.wrap(json.data).traverse(decode, Either)

    def decode_json_scalar(self, json: JsonScalar) -> Either[JsonError, A]:
        return Right(json.data)

    def decode_json_absent(self, json: JsonAbsent) -> Either[JsonError, A]:
        return Left(json.data)


decode = dispatch_alg(Decode(), Json, 'decode_')


@do(Either[JsonError, A])
def decode_json(data: str) -> Do:
    json = yield parse_json(data).lmap(lambda e: JsonError(data, e))
    yield decode(json)


@do(Either[JsonError, A])
def decode_json_type_json(json: Json, tpe: Type[A]) -> Do:
    dec = yield Decoder.e(tpe).lmap(L(JsonError)(json, _))
    yield dec.decode(tpe, json)


@do(Either[JsonError, A])
def decode_json_type(data: str, tpe: Type[A]) -> Do:
    json = yield parse_json(data).lmap(lambda e: JsonError(data, e))
    yield decode_json_type_json(json, tpe)


__all__ = ('Decoder', 'decode_json_object', 'decode_json', 'decode', 'decode_json_type', 'decode_json_type_json')



from typing import Type, TypeVar

A = TypeVar('A')


def qualified_type(tpe: Type[A]) -> str:
    return tpe.__name__ if tpe.__module__ == 'builtins' else f'{tpe.__module__}.{tpe.__name__}'

__all__ = ('qualified_type',)



from typing import Type, TypeVar, Collection, Mapping
from numbers import Number
from uuid import UUID

from amino.json.decoder import Decoder, decode, decode_json_type_json
from amino import Maybe, Either, List, Lists, Left, Boolean, Try, Map, Right, Nothing, Just, Path, do, Do
from amino.json.data import JsonError, Json, JsonObject

A = TypeVar('A')
B = TypeVar('B')


class StringDecoder(Decoder, tpe=str):

    def decode(self, tpe: Type[str], data: Json) -> Either[JsonError, str]:
        return data.scalar.e(f'invalid type for `str`: {data}', data.data)


class NumberDecoder(Decoder, sub=Number):

    def decode(self, tpe: Type[int], data: Json) -> Either[JsonError, int]:
        return data.scalar.e(f'invalid type for `int`: {data}', data.data)


class MapDecoder(Decoder, sub=Mapping):

    def decode(self, tpe: Type[Mapping], data: Json) -> Either[JsonError, Map[str, A]]:
        def dec() -> Either[JsonError, Map[str, A]]:
            return Map(data.data).traverse(decode, Either)
        return data.object.c(dec, lambda: Left(f'invalid type for `Map`: {data}'))


class ListDecoder(Decoder, sub=Collection):

    def decode(self, tpe: Type[Collection], data: Json) -> Either[JsonError, List[A]]:
        def dec() -> Either[JsonError, List[A]]:
            return Lists.wrap(data.data).traverse(decode, Either)
        return data.array.c(dec, lambda: Left(f'invalid type for `List`: {data}'))


def maybe_from_object(data: JsonObject, inner: Maybe[Type[A]]) -> Either[JsonError, Maybe[A]]:
    return (
        decode(data) / Just
        if data.has_type else
        inner.traverse(lambda a: decode_json_type_json(data, a), Either)
    )


class MaybeDecoder(Decoder, tpe=Maybe):

    def decode(self, tpe: Type[Maybe], data: Json) -> Either[JsonError, Maybe[A]]:
        inner = Lists.wrap(tpe.__args__).head
        return (
            Right(Nothing)
            if data.absent else
            maybe_from_object(data, inner)
            if data.object else
            decode(data) / Just
            if data.array else
            data.scalar.e(f'invalid type for `Maybe`: {data}', Maybe.check(data.data))
        )


@do(Either[JsonError, Either[A, B]])
def either_from_object(data: JsonObject, ltype: Type[A], rtype: Type[B]) -> Do:
    value = yield data.field('value')
    decoded = yield decode(value)
    tpe = yield data.tpe
    yield Right(tpe(decoded))


class EitherDecoder(Decoder, tpe=Either):

    @do(Either[JsonError, Either[A, B]])
    def decode(self, tpe: Type[Either], data: Json) -> Do:
        err = JsonError(data, f'too few types in Either')
        ltype, rtype = yield Lists.wrap(tpe.__args__).lift_all(0, 1).to_either(err)
        yield (
            either_from_object(data, ltype, rtype)
            if data.object else
            Left(JsonError(data, f'invalid type for `Either`'))
        )


class BooleanDecoder(Decoder, tpe=Boolean):

    def decode(self, tpe: Type[Boolean], data: Json) -> Either[JsonError, Boolean]:
        return data.scalar.e(f'invalid type for `Boolean`: {data}', Boolean(data.data))


class UUIDDecoder(Decoder, tpe=UUID):

    def decode(self, tpe: Type[UUID], data: Json) -> Either[JsonError, UUID]:
        return data.scalar.flat_e(f'invalid type for `UUID`: {data}', Try(UUID, data.data))


class PathDecoder(Decoder, tpe=Path):

    def decode(self, tpe: Type[Path], data: Json) -> Either[JsonError, Path]:
        return data.scalar.flat_e(f'invalid type for `Path`: {data}', Try(Path, data.data))


__all__ = ('MaybeDecoder', 'StringDecoder', 'NumberDecoder', 'ListDecoder', 'BooleanDecoder', 'MapDecoder',
           'PathDecoder')



import abc
from typing import TypeVar, Union, Generic, Generator
from numbers import Number

from amino import Map, List, _, Either, Right, Left, Boolean
from amino.util.string import ToStr
from amino.algebra import Algebra
from amino.do import tdo

A = TypeVar('A')
tpe_key = '__type__'


class JsonError(ToStr):

    def __init__(self, data: str, error: Union[str, Exception]) -> None:
        self.data = data
        self.error = error

    def _arg_desc(self) -> List[str]:
        return List(self.data, str(self.error))

    @property
    def exception(self) -> Exception:
        return self.error if isinstance(self.error, Exception) else Exception(self.error)


class Json(Generic[A], Algebra, base=True):

    def __init__(self, data: A) -> None:
        self.data = data

    @abc.abstractproperty
    def native(self) -> Union[dict, list, str, Number, None]:
        ...

    @abc.abstractmethod
    def field(self, key: str) -> Either[JsonError, 'Json']:
        ...

    def _arg_desc(self) -> List[str]:
        return List(str(self.data))

    @property
    def scalar(self) -> Boolean:
        return Boolean.isinstance(self, JsonScalar)

    @property
    def array(self) -> Boolean:
        return Boolean.isinstance(self, JsonArray)

    @property
    def object(self) -> Boolean:
        return Boolean.isinstance(self, JsonObject)

    @property
    def absent(self) -> Boolean:
        return Boolean.isinstance(self, JsonAbsent)

    def error(self, msg: Union[str, Exception]) -> JsonError:
        return JsonError(self.data, msg)


class JsonObject(Json[Map[str, Json]]):

    @property
    def native(self) -> Union[dict, list, str, Number, None]:
        return self.data.valmap(_.native)

    @property
    @tdo(Either[JsonError, type])
    def tpe(self) -> Generator:
        jtpe = yield self.data.lift(tpe_key).to_either(self.error(f'no `{tpe_key}` field in json object'))
        tpe_s = yield (
            Right(jtpe.data)
            if isinstance(jtpe, JsonScalar) else
            Left(self.error('invalid type for `{type_key}`: {jtpe}'))
        )
        yield Either.import_path(tpe_s).lmap(self.error)

    def field(self, key: str) -> Either[JsonError, Json]:
        return Right(self.data.lift(key) | JsonAbsent(self.error(f'no field `{key}`')))

    @property
    def has_type(self) -> Boolean:
        return self.data.contains(tpe_key)


class JsonArray(Json[List[Json]]):

    @property
    def native(self) -> Union[dict, list, str, Number, None]:
        return self.data.map(_.native)

    def field(self, key: str) -> Either[JsonError, Json]:
        return Left(self.error('JsonArray has no fields'))


class JsonScalar(Json[Union[str, Number, None]]):

    @property
    def native(self) -> Union[dict, list, str, Number, None]:
        return self.data

    def field(self, key: str) -> Either[JsonError, Json]:
        return Left(self.error('JsonScalar has no fields'))


class JsonAbsent(Json[JsonError]):

    @property
    def native(self) -> Union[dict, list, str, Number, None]:
        return None

    def field(self, key: str) -> Either[JsonError, Json]:
        return Right(self)


__all__ = ('JsonError', 'Json', 'JsonObject', 'JsonArray', 'JsonScalar', 'JsonAbsent')



import abc
import time
import traceback
import inspect
import typing
from typing import Callable, TypeVar, Generic, Any, Union, Tuple, Awaitable, Generator

from amino import Either, Right, Left, Maybe, List, __, Just, _, Lists, L, Nothing, options
from amino.eval import Eval
from amino.tc.base import Implicits, ImplicitsMeta
from amino.logging import log
from amino.util.fun import lambda_str, format_funcall
from amino.util.exception import format_exception, sanitize_tb
from amino.util.string import ToStr
from amino.do import do
from amino.func import tailrec

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')


class IOException(Exception):
    remove_pkgs = List('amino', 'fn')

    def __init__(self, f, stack, cause) -> None:
        self.f = f
        self.stack = List.wrap(stack)
        self.cause = cause

    @property
    def location(self):
        files = List('io', 'anon', 'instances/io', 'tc/base')
        def filt(entry, name):
            return entry.filename.endswith('/amino/{}.py'.format(name))
        stack = self.stack.filter_not(lambda a: files.exists(L(filt)(a, _)))
        pred = (lambda a: not IOException.remove_pkgs
                .exists(lambda b: '/{}/'.format(b) in a.filename))
        return stack.find(pred)

    @property
    def format_stack(self) -> List[str]:
        rev = self.stack.reversed
        def remove_recursion(i):
            pre = rev[:i + 1]
            post = rev[i:].drop_while(__.filename.endswith('/amino/io.py'))
            return pre + post
        def remove_internal():
            start = rev.index_where(_.function == 'unsafe_perform_sync')
            return start / remove_recursion | rev
        frames = (self.location.to_list if IO.stack_only_location else remove_internal())
        data = frames / (lambda a: a[1:-2] + tuple(a[-2]))
        return sanitize_tb(Lists.wrap(traceback.format_list(list(data))))

    @property
    def lines(self) -> List[str]:
        cause = format_exception(self.cause)
        suf1 = '' if self.stack.empty else ' at:'
        tb1 = (List() if self.stack.empty else self.format_stack)
        return tb1.cons(f'IO exception{suf1}').cat('Cause:') + cause + List(
            '',
            'Callback:',
            f'  {self.f}'
        )

    def __str__(self):
        return self.lines.join_lines


class IOMeta(ImplicitsMeta):

    @property
    def zero(self):
        return IO.now(None)


def safe_fmt(f: Callable[..., Any], a: tuple, kw: dict) -> Eval[str]:
    def s() -> str:
        try:
            return format_funcall(f, Lists.wrap(a), kw)
        except Exception as e:
            return str(f)
            log.error(str(e))
    return Eval.later(s)


class IO(Generic[A], Implicits, ToStr, implicits=True, metaclass=IOMeta):
    debug = options.io_debug.exists
    stack_only_location = True

    @staticmethod
    def delay(f: Callable[..., A], *a: Any, **kw: Any) -> 'IO[A]':
        return Suspend(L(f)(*a, **kw) >> Pure, safe_fmt(f, a, kw))

    @staticmethod
    def suspend(f: Callable[..., 'IO[A]'], *a: Any, **kw: Any) -> 'IO[A]':
        return Suspend(L(f)(*a, **kw), safe_fmt(f, a, kw))

    @staticmethod
    def call(f: Callable[..., A], *a, **kw):
        return IO.delay(f, *a, **kw)

    @staticmethod
    def now(a: A) -> 'IO[A]':
        return Pure(a)

    @staticmethod
    def pure(a: A) -> 'IO[A]':
        return Pure(a)

    @staticmethod
    def just(a: A) -> 'IO[Maybe[A]]':
        return IO.now(Just(a))

    @staticmethod
    def failed(err: str) -> 'IO[A]':
        def fail():
            raise Exception(err)
        return IO.suspend(fail)

    @staticmethod
    def from_either(a: Either[Any, A]) -> 'IO[A]':
        return a.cata(IO.failed, IO.now)

    @staticmethod
    def from_maybe(a: Maybe[A], error: str) -> 'IO[A]':
        return a / IO.now | IO.failed(error)

    def __init__(self) -> None:
        self.stack = inspect.stack() if IO.debug else []

    @abc.abstractmethod
    def _flat_map(self, f: Callable[[A], 'IO[B]'], ts: Eval[str], fs: Eval[str]) -> 'IO[B]':
        ...

    @abc.abstractmethod
    def _step(self) -> 'IO[A]':
        ...

    @abc.abstractmethod
    def lambda_str(self) -> Eval[str]:
        ...

    def _arg_desc(self) -> List[str]:
        return List(self.lambda_str()._value())

    def step(self) -> Union[A, 'IO[A]']:
        try:
            return self._step_timed() if IO.debug else self._step()
        except IOException as e:
            raise e
        except Exception as e:
            raise IOException(self.lambda_str()._value(), self.stack, e)

    def run(self) -> A:
        @tailrec
        def run(t: Union[A, 'IO[A]']) -> Union[Tuple[bool, A], Tuple[bool, Tuple[Union[A, 'IO[A]']]]]:
            if isinstance(t, Pure):
                return True, (t.value,)
            elif isinstance(t, IO):
                return True, (t.step(),)
            else:
                return False, t
        return run(self)

    def flat_map(self, f: Callable[[A], 'IO[B]'], ts: Maybe[Eval[str]]=Nothing, fs: Maybe[Eval[str]]=Nothing
                 ) -> 'IO[B]':
        ts1 = ts | self.lambda_str
        fs1 = fs | Eval.later(lambda: f'flat_map({lambda_str(f)})')
        return self._flat_map(f, ts1, fs1)

    def _step_timed(self) -> Union[A, 'IO[A]']:
        start = time.time()
        v = self._step()
        dur = time.time() - start
        if dur > 0.1:
            log.debug2(lambda: 'IO {} took {:.4f}s'.format(self.string(), dur))
        return v

    def _attempt(self) -> Either[IOException, A]:
        try:
            return Right(self.run())
        except IOException as e:
            return Left(e)

    def attempt_run(self) -> Either[IOException, A]:
        return self._attempt()

    @property
    def attempt(self) -> Either[IOException, A]:
        return self._attempt()

    @property
    def fatal(self) -> A:
        return self.attempt.get_or_raise

    def and_then(self, nxt: 'IO[B]') -> 'IO[B]':
        fs = Eval.later(lambda: 'and_then({nxt.lambda_str()})')
        return self.flat_map(lambda a: nxt, fs=Just(fs))

    __add__ = and_then

    def join_maybe(self, err: str) -> 'IO[B]':
        def f(a: Union[A, Maybe[B]]) -> 'IO[B]':
            return (
                IO.from_maybe(a, err)
                if isinstance(a, Maybe) else
                IO.failed(f'`IO.join_maybe` called on {self}')
            )
        return self.flat_map(f, fs=Just(Eval.later('join_maybe({})'.format, err)))

    @property
    def join_either(self) -> 'IO[B]':
        def f(a: Union[A, Either[C, B]]) -> 'IO[B]':
            return (
                IO.from_either(a)
                if isinstance(a, Either) else
                IO.failed(f'`IO.join_either` called on {self}')
            )
        return self.flat_map(f, fs=Just(Eval.now('join_either')))

    def with_stack(self, s: typing.List[inspect.FrameInfo]) -> 'IO[A]':
        self.stack = s
        return self

    def recover(self, f: Callable[[IOException], B]) -> 'IO[B]':
        return IO.delay(lambda: self.attempt).map(__.value_or(f))

    @do('IO[A]')
    def ensure(self, f: Callable[[Either[IOException, A]], 'IO[None]']) -> Generator:
        result = yield IO.delay(lambda: self.attempt)
        yield f(result)
        yield IO.from_either(result)

    @property
    def coro(self) -> Awaitable[Either[IOException, A]]:
        async def coro() -> Either[IOException, A]:
            return self.attempt
        return coro()


class Suspend(Generic[A], IO[A]):

    def __init__(self, thunk: Callable[[], IO[A]], string: Eval[str]) -> None:
        super().__init__()
        self.thunk = thunk
        self.string = string

    def lambda_str(self) -> Eval[str]:
        return self.string

    def _step(self) -> IO[A]:
        return self.thunk().with_stack(self.stack)

    def _flat_map(self, f: Callable[[A], IO[B]], ts: Eval[str], fs: Eval[str]) -> IO[B]:
        return BindSuspend(self.thunk, f, ts, fs)


class BindSuspend(Generic[A], IO[A]):

    def __init__(self, thunk: Callable[[], IO[A]], f: Callable, ts: Eval[str], fs: Eval[str]) -> None:
        super().__init__()
        self.thunk = thunk
        self.f = f
        self.ts = ts
        self.fs = fs

    def lambda_str(self) -> Eval[str]:
        return (self.ts & self.fs).map2('{}.{}'.format)

    def _step(self) -> IO[A]:
        return (
            self.thunk()
            .flat_map(self.f, fs=Just(self.fs))
            .with_stack(self.stack)
        )

    def _flat_map(self, f: Callable[[A], IO[B]], ts: Eval[str], fs: Eval[str]) -> IO[B]:
        bs = L(BindSuspend)(self.thunk, lambda a: self.f(a).flat_map(f, Just(ts), Just(fs)), ts, fs)
        return Suspend(bs, (ts & fs).map2('{}.{}'.format))


class Pure(Generic[A], IO[A]):

    def __init__(self, value: A) -> None:
        super().__init__()
        self.value = value

    def lambda_str(self) -> Eval[str]:
        return Eval.later(lambda: f'Pure({self.value})')

    def _arg_desc(self) -> List[str]:
        return List(str(self.value))

    def _step(self) -> IO[A]:
        return self

    def _flat_map(self, f: Callable[[A], IO[B]], ts: Eval[str], fs: Eval[str]) -> IO[B]:
        return Suspend(L(f)(self.value), (ts & fs).map2('{}.{}'.format))


def Try(f: Callable[..., A], *a: Any, **kw: Any) -> Either[Exception, A]:
    try:
        return Right(f(*a, **kw))
    except Exception as e:
        return Left(e)


def io(fun: Callable[..., A]) -> Callable[..., IO[A]]:
    def dec(*a: Any, **kw: Any) -> IO[A]:
        return IO.delay(fun, *a, **kw)
    return dec

__all__ = ('IO', 'Try', 'io')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0x40340b99

# Compiled with Coconut version 1.3.0 [Dead Parrot]

# Coconut Header: -------------------------------------------------------------

from __future__ import generator_stop
import sys as _coconut_sys, os.path as _coconut_os_path
_coconut_file_path = _coconut_os_path.dirname(_coconut_os_path.abspath(__file__))
_coconut_sys.path.insert(0, _coconut_file_path)
from __coconut__ import _coconut, _coconut_NamedTuple, _coconut_MatchError, _coconut_tail_call, _coconut_tco, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_pipe, _coconut_star_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial
from __coconut__ import *
_coconut_sys.path.remove(_coconut_file_path)

# Compiled Coconut: -----------------------------------------------------------

import json
from typing import Any

from amino import Either
from amino import Left
from amino import Lists
from amino import Map
from amino import Try
from amino.json.data import Json
from amino.json.data import JsonArray
from amino.json.data import JsonScalar
from amino.json.data import JsonObject


def to_json(a: 'Any') -> 'Json':
    _coconut_match_to = a
    _coconut_match_check = False
    if _coconut.isinstance(_coconut_match_to, (list, tuple)):
        a = _coconut_match_to
        _coconut_match_check = True
    if _coconut_match_check:
        result = JsonArray(Lists.wrap(a) / to_json)
    if not _coconut_match_check:
        if _coconut.isinstance(_coconut_match_to, dict):
            a = _coconut_match_to
            _coconut_match_check = True
        if _coconut_match_check:
            result = JsonObject(Map(a).valmap(to_json))
    if not _coconut_match_check:
        a = _coconut_match_to
        _coconut_match_check = True
        if _coconut_match_check:
            result = JsonScalar(a)
    return result


def parse_json(payload: 'str') -> 'Either[str, Json]':
    return Try(json.loads, payload) / to_json



from typing import Union, Collection, TypeVar, Mapping, Any, List as TList
from numbers import Number
from uuid import UUID

from amino import Either, List, L, _, Right, Lists, Maybe, Path, Map, Boolean, do, Do
from amino.json.encoder import Encoder, encode_json, json_object_with_type
from amino.json.data import JsonError, Json, JsonArray, JsonScalar, JsonObject

A = TypeVar('A')
B = TypeVar('B')


class ScalarEncoder(Encoder[Union[Number, str, None]], pred=L(issubclass)(_, (Number, str, type(None)))):

    def encode(self, a: Union[Number, str, None]) -> Either[JsonError, Json]:
        return Right(JsonScalar(a))


class MapEncoder(Encoder[List], pred=L(issubclass)(_, Mapping)):

    def encode(self, a: Map[str, Any]) -> Either[JsonError, Json]:
        return Map(a).traverse(encode_json, Either) / JsonObject


class ListEncoder(Encoder[List], pred=L(issubclass)(_, TList)):

    def encode(self, a: TList) -> Either[JsonError, Json]:
        return Lists.wrap(a).traverse(encode_json, Either) / JsonArray


class MaybeEncoder(Encoder[Maybe], tpe=Maybe):

    def encode(self, a: Maybe[A]) -> Either[JsonError, Json]:
        return Right(JsonScalar(a | None))


class EitherEncoder(Encoder[Either], tpe=Either):

    @do(Either[JsonError, Json])
    def encode(self, a: Either[B, A]) -> Do:
        json = yield encode_json(a.value)
        yield Right(json_object_with_type(Map(value=json), type(a)))


class UUIDEncoder(Encoder[UUID], tpe=UUID):

    def encode(self, a: UUID) -> Either[JsonError, Json]:
        return Right(JsonScalar(str(a)))


class PathEncoder(Encoder[Path], tpe=Path):

    def encode(self, a: Path) -> Either[JsonError, Json]:
        return Right(JsonScalar(str(a)))


class TypeEncoder(Encoder[type], tpe=type):

    def encode(self, a: type) -> Either[JsonError, Json]:
        return Right(json_object_with_type(Map(), a))


class BooleanEncoder(Encoder[Boolean], tpe=Boolean):

    def encode(self, a: Boolean) -> Either[JsonError, Json]:
        return Right(JsonScalar(a.value))


__all__ = ('ListEncoder', 'ScalarEncoder', 'MaybeEncoder', 'UUIDEncoder', 'PathEncoder', 'TypeEncoder', 'MapEncoder')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0xc65d39be

# Compiled with Coconut version 1.3.0 [Dead Parrot]

# Coconut Header: -------------------------------------------------------------

from __future__ import generator_stop
import sys as _coconut_sys, os.path as _coconut_os_path
_coconut_file_path = _coconut_os_path.dirname(_coconut_os_path.abspath(__file__))
_coconut_sys.path.insert(0, _coconut_file_path)
from __coconut__ import _coconut, _coconut_NamedTuple, _coconut_MatchError, _coconut_tail_call, _coconut_tco, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_pipe, _coconut_star_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial
from __coconut__ import *
_coconut_sys.path.remove(_coconut_file_path)

# Compiled Coconut: -----------------------------------------------------------

import abc
import importlib
from typing import TypeVar
from typing import Generic
from typing import Union
from typing import Any
from typing import cast
from typing import Iterator
from typing import Type
from typing import Generator
from types import ModuleType
from pathlib import Path

import amino
from amino import boolean
from amino.func import I
from amino.tc.base import F
from amino.util.mod import unsafe_import_name
from amino.tc.monoid import Monoid
from amino.util.string import ToStr
from amino.do import do
from amino.util.exception import format_exception

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
D = TypeVar('D')
E = TypeVar('E', bound=Exception)


class ImportFailure(ToStr):

    @abc.abstractproperty
    def expand(self) -> 'amino.List[str]':
        ...


class ImportException(ImportFailure):

    def __init__(self, desc: 'str', exc: 'Exception') -> 'None':
        self.desc = desc
        self.exc = exc

    @_coconut_tco
    def _arg_desc(self) -> 'amino.List[str]':
        return _coconut_tail_call(amino.List, self.desc, str(self.exc))

    @property
    @_coconut_tco
    def expand(self) -> 'amino.List[str]':
        return _coconut_tail_call(format_exception(self.exc).cons, self.desc)


class InvalidLocator(ImportFailure):

    def __init__(self, msg: 'str') -> 'None':
        self.msg = msg

    @_coconut_tco
    def _arg_desc(self) -> 'amino.List[str]':
        return _coconut_tail_call(amino.List, self.msg)

    @property
    @_coconut_tco
    def expand(self) -> 'amino.List[str]':
        return _coconut_tail_call(amino.List, self.msg)


class Either(Generic[A, B], F[B], implicits=True):

    @staticmethod
    @_coconut_tco
    def import_name(mod: 'str', name: 'str') -> 'Either[ImportFailure, B]':
        try:
            value = unsafe_import_name(mod, name)
        except Exception as e:
            return _coconut_tail_call(Left, ImportException(f'{mod}.{name}', e))
        else:
            return Left(InvalidLocator(f'{mod} has no attribute {name}')) if value is None else Right(value)

    @staticmethod
    def import_path(path: 'str') -> 'Either[ImportFailure, B]':
        from amino.list import List
        return (List.wrap(path.rsplit('.', 1)).lift_all(0, 1).to_either(InvalidLocator(f'invalid module path: {path}')).flat_map2(lambda a, b: Either.import_name(a, b).lmap(lambda c: ImportException(path, c))))

    @staticmethod
    @_coconut_tco
    def import_module(modname: 'str') -> 'Either[ImportFailure, ModuleType]':
        try:
            mod = importlib.import_module(modname)
        except Exception as e:
            return _coconut_tail_call(Left, ImportException(modname, e))
        else:
            return _coconut_tail_call(Right, mod)

    @staticmethod
    @_coconut_tco
    def import_file(path: 'Path') -> 'Either[ImportFailure, ModuleType]':
        from amino.maybe import Maybe
        def step2(spec: 'importlib._bootstrap.ModuleSpec') -> 'ModuleType':
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            return module
        try:
            module = Maybe.check(importlib.util.spec_from_file_location('temp', str(path))) / step2
        except Exception as e:
            return _coconut_tail_call(Left, ImportException(str(path), e))
        else:
            return _coconut_tail_call(module.to_either, InvalidLocator(f'failed to import `{path}`'))

    @staticmethod
    @do('Either[ImportFailure, B]')
    def import_from_file(path: 'Path', name: 'str') -> 'Generator':
        module = yield Either.import_file(path)
        attr = getattr(module, name, None)
        yield (Left(InvalidLocator(f'{path} has no attribute {name}')) if attr is None else Right(attr))

    @staticmethod
    @_coconut_tco
    def catch(f: '_coconut.typing.Callable[[], B]', exc: 'Type[E]') -> 'Either[E, B]':
        try:
            return Right(f())
        except exc as e:
            return _coconut_tail_call(Left, e)

    @staticmethod
    @do('Either[ImportFailure, amino.List[Any]]')
    def exports(modpath: 'str') -> 'Generator':
        from amino.list import Lists
        exports = yield Either.import_name(modpath, '__all__')
        yield Lists.wrap(exports).traverse(lambda a: Either.import_name(modpath, a), Either)

    @property
    @_coconut_tco
    def is_right(self) -> 'amino.Boolean':
        return _coconut_tail_call(boolean.Boolean, isinstance(self, Right))

    @property
    @_coconut_tco
    def is_left(self) -> 'amino.Boolean':
        return _coconut_tail_call(boolean.Boolean, isinstance(self, Left))

    @property
    @_coconut_tco
    def __left_value(self) -> 'A':
        return _coconut_tail_call(cast, A, self.value)

    @property
    @_coconut_tco
    def __right_value(self) -> 'B':
        return _coconut_tail_call(cast, B, self.value)

    def leffect(self, f: '_coconut.typing.Callable[[A], Any]') -> 'Either[A, B]':
        if self.is_left:
            f(self.__left_value)
        return self

    def bieffect(self, l: '_coconut.typing.Callable[[A], Any]', r: '_coconut.typing.Callable[[B], Any]') -> 'Either[A, B]':
        self.cata(l, r)
        return self

    def cata(self, fl: '_coconut.typing.Callable[[A], C]', fr: '_coconut.typing.Callable[[B], C]') -> 'C':
        return fl(self.__left_value) if self.is_left else fr(self.__right_value)

    def bimap(self, fl: '_coconut.typing.Callable[[A], C]', fr: '_coconut.typing.Callable[[B], D]') -> 'Either[C, D]':
        return Left(fl(self.__left_value)) if self.is_left else Right(fr(self.__right_value))

    @_coconut_tco
    def recover_with(self, f: '_coconut.typing.Callable[[A], Either[C, B]]') -> 'Either[C, B]':
        return _coconut_tail_call(self.cata, f, Right)

    @_coconut_tco
    def right_or_map(self, f: '_coconut.typing.Callable[[A], C]') -> 'C':
        return _coconut_tail_call(self.cata, f, I)

    @_coconut_tco
    def value_or(self, f: '_coconut.typing.Callable[[A], B]') -> 'B':
        return _coconut_tail_call(self.cata, f, I)

    @_coconut_tco
    def left_or(self, f: '_coconut.typing.Callable[[B], A]') -> 'A':
        return _coconut_tail_call(self.cata, I, f)

    @_coconut_tco
    def left_or_map(self, f: '_coconut.typing.Callable[[B], A]') -> 'A':
        return _coconut_tail_call(self.cata, I, f)

    @property
    @_coconut_tco
    def ljoin(self) -> 'Either[A, C]':
        return _coconut_tail_call(self.right_or_map, Left)

    @_coconut_tco
    def __str__(self) -> 'str':
        return _coconut_tail_call('{}({})'.format, self.__class__.__name__, self.value)

    @_coconut_tco
    def __repr__(self) -> 'str':
        return _coconut_tail_call('{}({!r})'.format, self.__class__.__name__, self.value)

    @property
    def to_list(self) -> 'amino.List[B]':
        return self.to_maybe.to_list

    def lmap(self, f: '_coconut.typing.Callable[[A], C]') -> 'Either[C, B]':
        return cast(Either, Left(f(self.__left_value))) if self.is_left else cast(Either, Right(self.__right_value))

    @_coconut_tco
    def get_or_raise(self) -> 'B':
        def fail(err: 'A') -> 'B':
            raise err if isinstance(err, Exception) else Exception(err)
        return _coconut_tail_call(self.cata, fail, I)

    @property
    def fatal(self) -> 'B':
        return self.get_or_raise

    @_coconut_tco
    def __iter__(self) -> 'Iterator[B]':
        return _coconut_tail_call(iter, self.to_list)

    @property
    @_coconut_tco
    def swap(self) -> 'Either[B, A]':
        return _coconut_tail_call(self.cata, Right, Left)

    @property
    def json_repr(self) -> 'B':
        return self.to_maybe.json_repr

    @_coconut_tco
    def accum_error(self, b: 'Either[A, C]') -> 'Either[A, C]':
        return _coconut_tail_call(self.accum_error_f, lambda: b)

    def accum_error_f(self, f: '_coconut.typing.Callable[[], Either[A, C]]') -> 'Either[A, C]':
        @_coconut_tco
        def acc(v: 'A') -> 'None':
            return _coconut_tail_call(Monoid.fatal(type(v)).combine, self.__left_value, v)
        return f().lmap(acc) if self.is_left else self

    def filter_with(self, f: '_coconut.typing.Callable[[B], bool]', g: '_coconut.typing.Callable[[B], C]') -> 'Either[C, B]':
        return self // (lambda a: Right(a) if f(a) else Left(g(a)))

    @_coconut_tco
    def left_contains(self, a: 'A') -> 'boolean.Boolean':
        return _coconut_tail_call(boolean.Boolean, self.is_left and self.__left_value == a)


class Right(_coconut.collections.namedtuple("Right", "value"), Generic[A, B], Either[A, B]):
    __slots__ = ()
    __ne__ = _coconut.object.__ne__
    def __eq__(self, other: 'Any') -> 'bool':
        return isinstance(other, Right) and self._Either__right_value == other._Either__right_value


class Left(_coconut.collections.namedtuple("Left", "value"), Generic[A, B], Either[A, B]):
    __slots__ = ()
    __ne__ = _coconut.object.__ne__
    def __eq__(self, other: 'Any') -> 'bool':
        return isinstance(other, Left) and self._Either__left_value == other._Either__left_value

__all__ = ('Either', 'Left', 'Right', 'ImportFailure', 'ImportException', 'InvalidLocator')
